"""Provides a widget for showing a PEP's details."""

##############################################################################
# Python imports.
from datetime import date, datetime
from functools import singledispatchmethod
from typing import Final, Sequence
from webbrowser import open as visit_url

##############################################################################
# Humanize imports
from humanize import naturaltime

##############################################################################
# Rich imports.
from rich.emoji import Emoji

##############################################################################
# Textual imports.
from textual import _widget_navigation, on
from textual.app import ComposeResult
from textual.containers import Vertical, VerticalScroll
from textual.events import DescendantBlur, DescendantFocus
from textual.reactive import var
from textual.types import Direction
from textual.widgets import Label
from textual.widgets.option_list import Option

##############################################################################
# Local imports.
from ...peps import PEP, PEPStatus, PEPType, PostHistory
from ..messages import (
    GotoPEP,
    ShowAuthor,
    ShowPythonVersion,
    ShowStatus,
    ShowType,
    VisitPEP,
)
from .extended_option_list import OptionListEx


##############################################################################
def date_display(display_date: date) -> str:
    """Format a date for display.

    Args:
        display_date: The date to format for display.

    Returns:
        The date formatted for display.
    """
    return f"{display_date} ({naturaltime(datetime.combine(display_date, datetime.min.time()))})"


##############################################################################
class Field(Vertical):
    """A container class that gives a widget a title."""

    DEFAULT_CSS = """
    Field {
        height: auto;
        margin-bottom: 1;
        &.hidden {
            display: none;
        }

        & > #_field_title {
            color: $text-accent;
            width: 1fr;
            background: $boost;
        }
    }
    """

    def __init__(self, title: str) -> None:
        """Initialise the widget.

        Args:
            title: The title to give the widget.
        """
        super().__init__(classes="hidden")
        self.compose_add_child(Label(title, id="_field_title"))


##############################################################################
class Value(Label):
    """A label that will hide itself if empty or `None`."""

    def show(self, text: str | int | None) -> None:
        """Show the given text, or possibly hide.

        Args:
            text: The text to show, or `None`.
        """
        if self.parent is None:
            return
        if isinstance(text, int):
            text = str(text)
        self.parent.set_class(not bool(text), "hidden")
        if text:
            self.update(text)


##############################################################################
class Item(Option):
    """Type of an item that goes in a list."""

    def select(self, parent: OptionListEx) -> None:
        """Perform the selection action for the item.

        Args:
            parent: The parent list for the item.
        """


##############################################################################
class StatusItem(Item):
    """Type of an item that is a PEP status."""

    def __init__(self, status: PEPStatus) -> None:
        """Initialise the object.

        Args:
            status: The status.
        """
        self._status: PEPStatus = status
        super().__init__(status)

    def select(self, parent: OptionListEx) -> None:
        """Perform the selection action for the item.

        Args:
            parent: The parent list for the item.
        """
        parent.post_message(ShowStatus(self._status))


##############################################################################
class TypeItem(Item):
    """Type of an item that is a PEP type."""

    def __init__(self, pep_type: PEPType) -> None:
        """Initialise the object.

        Args:
            pep_type: The PEP type.
        """
        self._type: PEPType = pep_type
        super().__init__(pep_type)

    def select(self, parent: OptionListEx) -> None:
        """Perform the selection action for the item.

        Args:
            parent: The parent list for the item.
        """
        parent.post_message(ShowType(self._type))


##############################################################################
class PEPItem(Item):
    """Type of an item that is a PEP number."""

    def __init__(self, pep: int) -> None:
        """Initialise the object.

        Args:
            pep: The number of the pep.
        """
        self._pep = pep
        super().__init__(f"PEP{pep}")

    def select(self, parent: OptionListEx) -> None:
        """Perform the selection action for the item.

        Args:
            parent: The parent list for the item.
        """
        parent.post_message(GotoPEP(self._pep))


##############################################################################
class AuthorItem(Item):
    """Type of an item that shows an author."""

    def __init__(self, author: str) -> None:
        """Initialise the object.

        Args:
            author: The author.
        """
        self._author = author
        super().__init__(author)

    def select(self, parent: OptionListEx) -> None:
        """Perform the selection action for the item.

        Args:
            parent: The parent list for the item.
        """
        parent.post_message(ShowAuthor(self._author))


##############################################################################
class PythonVersionItem(Item):
    """Type of an item that is a Python version."""

    def __init__(self, python_version: str) -> None:
        """Initialise the object.

        Args:
            python_version: The Python version.
        """
        self._version = python_version
        super().__init__(python_version)

    def select(self, parent: OptionListEx) -> None:
        """Perform the selection action for the item.

        Args:
            parent: The parent list for the item.
        """
        parent.post_message(ShowPythonVersion(self._version))


##############################################################################
class URLItem(Item):
    """Type of an item that shows a URL."""

    def __init__(self, url: str, title: str | None = None) -> None:
        """Initialise the object.

        Args:
            url: The URL.
            title: An optional title for the URL.
        """
        self._url = url
        super().__init__(title or url)

    def select(self, parent: OptionListEx) -> None:
        """Perform the selection action for the item.

        Args:
            parent: The parent list for the item.
        """
        visit_url(self._url)


##############################################################################
class PostItem(Item):
    """Type of an item that is a post."""

    LINK_ICON: Final[str] = Emoji.replace(":link:")
    """The icon to show if a date is linked."""

    def __init__(self, post: PostHistory) -> None:
        """Initialise the object.

        Args:
            post: The post.
        """
        self._post = post
        title = "Unknown"
        if post.date:
            title = (
                f"{date_display(post.date)}{' ' + self.LINK_ICON if post.url else ''}"
            )
        elif post.url:
            title = post.url
        super().__init__(title)

    def select(self, parent: OptionListEx) -> None:
        """Perform the selection action for the item.

        Args:
            parent: The parent list for the item.
        """
        if self._post.url:
            visit_url(self._post.url)
        else:
            parent.notify("There is no URL associated with this.", title="No URL")


##############################################################################
class ClickableValue(OptionListEx):
    """Show a value that the user can interact with."""

    DEFAULT_CSS = """
    ClickableValue, ClickableValue:focus {
        height: auto;
        width: auto;
        border: none;
        background: transparent !important;
        margin: 0;
        padding: 0;
        /* Stop a flash of unnecessary scrollbar. */
        scrollbar-size-vertical: 0;
        overflow-y: hidden;
    }
    """

    def __init__(self, id: str | None = None):
        """Initialise the object.

        Args:
            id: The ID of the object.
        """
        super().__init__(id=id)
        self._highlighted_memory: int | None = None
        """Holds the last highlighted item.

        This is to help `on_focus` and `on_blur`. Normally, as the user tabs
        through the widget, I want it to highlight and highlight; I'm doing
        this using focus and blur. However, when the user clicks away from
        the terminal this will also have the same effect. There I want them
        to be where they were when they comeback. So I use this to try and
        settle back in the right location.
        """

    @singledispatchmethod
    def show(self, values: Sequence[Item | None]) -> None:
        """Show a list of values.

        Args:
            values: The values to show.

        Notes:
            Any values that are empty will be removed. If the list is empty
            after this filter the field will be hidden.
        """
        self._highlighted_memory = None
        if self.parent is None:
            return
        self.clear_options().add_options([value for value in values if value])
        self.parent.set_class(not bool(self.option_count), "hidden")

    @show.register
    def _(self, values: Item | None) -> None:
        """Show a single value."""
        self.show([values])

    def on_focus(self) -> None:
        """Ensure the highlight appears when we get focus."""
        if self.highlighted is None and self.option_count:
            if (
                self._highlighted_memory is not None
                and self._highlighted_memory <= self.option_count
            ):
                self.highlighted = self._highlighted_memory
            else:
                self.highlighted = 0

    def on_blur(self) -> None:
        """Remove the highlight when we no longer have focus."""
        self._highlighted_memory = self.highlighted
        self.highlighted = None

    @on(OptionListEx.OptionSelected)
    def select(self, message: OptionListEx.OptionSelected) -> None:
        assert isinstance(message.option, Item)
        message.option.select(self)

    def _cursor_move(self, direction: Direction) -> int | None:
        """Work out the impact of an intended cursor move.

        Args:
            direction: The direction to move.

        Returns:
            The location to move the highlight to.
        """
        return _widget_navigation.find_next_enabled(
            self._options,
            anchor=self.highlighted,
            direction=direction,
        )

    def action_cursor_up(self) -> None:
        """Handle a cursor up action."""
        if (target := self._cursor_move(-1)) is None or self.highlighted is None:
            return
        if target >= self.highlighted:
            self.screen.focus_previous()
        else:
            self.highlighted = target

    def action_cursor_down(self) -> None:
        """Handle a cursor down action."""
        if (target := self._cursor_move(1)) is None or self.highlighted is None:
            return
        if target <= self.highlighted:
            self.screen.focus_next()
        else:
            self.highlighted = target


##############################################################################
class PEPDetails(VerticalScroll, can_focus=False):
    """A widget for showing details of a PEP."""

    HELP = """
    ## The Highlighted PEP's Details

    This panel shows the main details about the currently-highlighted PEP.
    `Tab` through the interactive fields and use them to apply more
    filtering or to visit external resources relating to the PEP.
    """

    pep: var[PEP | None] = var(None)
    """The PEP to show the details of."""

    BINDINGS = [("enter", "visit_pep")]

    def compose(self) -> ComposeResult:
        with Field("Title"):
            yield Value(id="title")
        with Field("Author"):
            yield ClickableValue(id="author")
        with Field("Sponsor"):
            yield Value(id="sponsor")
        with Field("Delegate"):
            yield Value(id="delegate")
        with Field("Discussions To"):
            yield Value(id="discussions_to")
        with Field("Status"):
            yield ClickableValue(id="status")
        with Field("Type"):
            yield ClickableValue(id="type")
        with Field("Topic"):
            yield Value(id="topic")
        with Field("Requires"):
            yield ClickableValue(id="requires")
        with Field("Replaces"):
            yield ClickableValue(id="replaces")
        with Field("Superseded By"):
            yield ClickableValue(id="superseded_by")
        with Field("Created"):
            yield Value(id="created")
        with Field("Python Version"):
            yield ClickableValue(id="python_versions")
        with Field("Post History"):
            yield ClickableValue(id="post_history")
        with Field("Resolution"):
            yield ClickableValue(id="resolution")
        with Field("URL"):
            yield ClickableValue(id="url")

    @on(DescendantBlur)
    @on(DescendantFocus)
    def _textual_5488_workaround(self) -> None:
        """Workaround for https://github.com/Textualize/textual/issues/5488"""
        for widget in self.query(ClickableValue):
            widget._refresh_lines()

    def watch_pep(self) -> None:
        """React to the PEP being changed."""
        with self.app.batch_update():
            if self.pep is None:
                self.query(Field).set_class(True, "hidden")
            else:
                self.query_one("#title", Value).show(self.pep.title)
                self.query_one("#author", ClickableValue).show(
                    AuthorItem(author) for author in self.pep.authors
                )
                self.query_one("#sponsor", Value).show(self.pep.sponsor)
                self.query_one("#delegate", Value).show(self.pep.delegate)
                self.query_one("#discussions_to", Value).show(self.pep.discussions_to)
                self.query_one("#status", ClickableValue).show(
                    StatusItem(self.pep.status)
                )
                self.query_one("#type", ClickableValue).show(TypeItem(self.pep.type))
                self.query_one("#topic", Value).show(
                    (self.pep.topic or "").capitalize()
                )
                self.query_one("#requires", ClickableValue).show(
                    PEPItem(pep) for pep in self.pep.requires
                )
                self.query_one("#replaces", ClickableValue).show(
                    PEPItem(pep) for pep in self.pep.replaces
                )
                self.query_one("#superseded_by", ClickableValue).show(
                    None
                    if self.pep.superseded_by is None
                    else PEPItem(self.pep.superseded_by)
                )
                self.query_one("#created", Value).show(date_display(self.pep.created))
                self.query_one("#python_versions", ClickableValue).show(
                    PythonVersionItem(version) for version in self.pep.python_version
                )
                self.query_one("#post_history", ClickableValue).show(
                    PostItem(post) for post in self.pep.post_history
                )
                self.query_one("#resolution", ClickableValue).show(
                    None
                    if self.pep.resolution is None
                    else PostItem(self.pep.resolution)
                )
                self.query_one("#url", ClickableValue).show(URLItem(self.pep.url))

    def action_visit_pep(self) -> None:
        """Action that visits the current PEP."""
        if self.pep is not None:
            self.post_message(VisitPEP(self.pep))


### pep_details.py ends here



"""The help screen for the application."""

##############################################################################
# Python imports.
from inspect import cleandoc
from operator import methodcaller
from typing import Any
from webbrowser import open as open_url

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Vertical, VerticalScroll
from textual.dom import DOMNode
from textual.screen import ModalScreen, Screen
from textual.widgets import Button, Markdown

##############################################################################
# Local imports.
from ... import __version__
from ..commands import Command

##############################################################################
# The help text.
HELP = f"""\
# Peplum v{__version__}

{{context_help}}

## About

`Peplum` is a terminal-based Python PEP lookup manager; it was created by
and is maintained by [Dave Pearson](https://www.davep.org/); it is Free
Software and can be [found on GitHub](https://github.com/davep/peplum).


## Licence

Peplum - The PEP lookup manager for the terminal.[EOL]
Copyright (C) 2025 Dave Pearson

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.
"""


##############################################################################
class HelpScreen(ModalScreen[None]):
    """The help screen."""

    CSS = """
    HelpScreen {
        align: center middle;

        &> Vertical {
            width: 75%;
            height: 90%;
            background: $panel;
            border: solid $border;
        }

        Markdown, MarkdownTable {
            padding: 0 1 0 1;
            background: transparent;
        }

        MarkdownH1 {
            padding: 1 0 1 0;
            background: $foreground 10%;
        }

        VerticalScroll {
            scrollbar-gutter: stable;
            scrollbar-background: $panel;
            scrollbar-background-hover: $panel;
            scrollbar-background-active: $panel;
        }

        Center {
            height: auto;
            width: 100%;
            border-top: solid $border;
        }
    }
    """

    BINDINGS = [("escape, f1", "close")]

    def __init__(self, help_for: Screen[Any]) -> None:
        """Initialise the help screen.

        Args:
            help_for: The screen to show the help for.
        """
        super().__init__()
        self._context_help = ""
        for node in (
            help_for.focused if help_for.focused is not None else help_for
        ).ancestors_with_self:
            if node.HELP is not None:
                self._context_help += f"\n\n{cleandoc(node.HELP)}"
            self._context_help += self.command_help(node)

    def _all_keys(self, command: Command) -> str:
        """Render all the keys for the given command.

        Args:
            command: The command to get all the keys for.

        Returns:
            A string listing all the keys for the command.
        """
        return ", ".join(
            self.app.get_key_display(Binding(key.strip(), ""))
            for key in command.binding().key.split(",")
        )

    def command_help(self, node: DOMNode) -> str:
        """Build help from the commands provided by a DOM node.

        Args:
            node: The node that might provide commands

        Returns:
            The help text.
        """
        if (commands := getattr(node, "COMMAND_MESSAGES", None)) is None:
            return ""
        keys = "| Command | Key | Description |\n| - | - | - |\n"
        for command in sorted(commands, key=methodcaller("command")):
            keys += f"| {command.command()} | {self._all_keys(command)} | {command.tooltip()} |\n"
        return f"\n\n{keys}"

    def compose(self) -> ComposeResult:
        """Compose the layout of the help screen."""
        with Vertical() as help_screen:
            help_screen.border_title = "Help"
            with VerticalScroll():
                yield Markdown(
                    HELP.replace("[EOL]", "  ").format(context_help=self._context_help)
                )
            with Center():
                yield Button("Okay [dim]\\[Esc]")

    @on(Button.Pressed)
    def action_close(self) -> None:
        """Close the help screen."""
        self.dismiss(None)

    @on(Markdown.LinkClicked)
    def visit(self, event: Markdown.LinkClicked) -> None:
        """Visit any link clicked in the help."""
        open_url(event.href)


### help.py ends here



"""PEP status filtering commands for the command palette."""

##############################################################################
# Local imports.
from ..data import PEPs
from ..messages import ShowStatus
from .commands_provider import CommandHit, CommandHits, CommandsProvider


##############################################################################
class StatusCommands(CommandsProvider):
    """A command palette provider related to statuses."""

    active_peps: PEPs | None = None
    """The currently-active collection of PEPs to get the statuses of."""

    @classmethod
    def prompt(cls) -> str:
        """The prompt for the command provider."""
        return (
            "Also search for PEPs of with status..."
            if cls.active_peps and cls.active_peps.is_filtered
            else "Search for PEPs with status..."
        )

    def commands(self) -> CommandHits:
        """Provide the status-based command data for the command palette.

        Yields:
            The commands for the command palette.
        """
        if self.active_peps is None:
            return
        help_prefix = "Also filter" if self.active_peps.is_filtered else "Filter"
        command_prefix = (
            "Also with status" if self.active_peps.is_filtered else "With status"
        )
        for status in self.active_peps.statuses:
            yield CommandHit(
                f"{command_prefix} {status.status}",
                f"{help_prefix} to PEPs with status {status.status} (narrows down to {status.count})",
                ShowStatus(status.status),
            )


### statuses.py ends here



"""Provides the command palette command provides for the application."""

##############################################################################
# Local imports.
from .authors import AuthorCommands
from .commands_provider import CommandsProvider
from .main import MainCommands
from .peps import PEPsCommands
from .python_versions import PythonVersionCommands
from .statuses import StatusCommands
from .types import TypeCommands

##############################################################################
# Exports.
__all__ = [
    "AuthorCommands",
    "CommandsProvider",
    "MainCommands",
    "PEPsCommands",
    "PythonVersionCommands",
    "StatusCommands",
    "TypeCommands",
]

### __init__.py ends here



"""Provides the main screen for the application."""

##############################################################################
# Python imports.
from dataclasses import dataclass
from json import dumps, loads
from webbrowser import open as visit_url

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.command import CommandPalette
from textual.containers import Horizontal
from textual.message import Message
from textual.reactive import var
from textual.screen import Screen
from textual.widgets import Footer, Header

##############################################################################
# Local imports.
from ... import __version__
from ...peps import API, PEP
from ..commands import (
    ChangeTheme,
    Command,
    Escape,
    FindPEP,
    Help,
    Quit,
    RedownloadPEPs,
    Search,
    SearchAuthor,
    SearchPythonVersion,
    SearchStatus,
    SearchType,
    ShowAll,
    SortByCreated,
    SortByNumber,
    SortByTitle,
    ToggleAuthorsSortOrder,
    TogglePEPDetails,
    TogglePythonVersionsSortOrder,
    ToggleStatusesSortOrder,
    ToggleTypesSortOrder,
)
from ..data import (
    Containing,
    PEPs,
    WithAuthor,
    WithPythonVersion,
    WithStatus,
    WithType,
    load_configuration,
    pep_data,
    update_configuration,
)
from ..messages import (
    GotoPEP,
    ShowAuthor,
    ShowPythonVersion,
    ShowStatus,
    ShowType,
    VisitPEP,
)
from ..providers import (
    AuthorCommands,
    CommandsProvider,
    MainCommands,
    PEPsCommands,
    PythonVersionCommands,
    StatusCommands,
    TypeCommands,
)
from ..widgets import Navigation, PEPDetails, PEPsView
from .help import HelpScreen
from .search_input import SearchInput


##############################################################################
class Main(Screen[None]):
    """The main screen for the application."""

    TITLE = f"Peplum v{__version__}"

    HELP = """
    ## Main application keys and commands

    The following keys and commands can be used anywhere here on the main screen.
    """

    DEFAULT_CSS = """
    Main {
        .panel {
            height: 1fr;
            border: none;
            border-left: round $border 50%;
            background: $surface;
            padding-right: 0;
            scrollbar-background: $surface;
            scrollbar-background-hover: $surface;
            scrollbar-background-active: $surface;
            &:focus, &:focus-within {
                border: none;
                border-left: round $border;
                background: $panel 80%;
                scrollbar-background: $panel;
                scrollbar-background-hover: $panel;
                scrollbar-background-active: $panel;
            }
            & > .option-list--option {
                padding: 0 1;
            }
        }

        Navigation {
            width: 2fr;
        }

        PEPsView {
            width: 8fr;
        }

        PEPDetails {
            width: 3fr;
            display: none;
        }

        &.details-visible {
            PEPsView {
                width: 5fr;
            }
            PEPDetails {
                display: block;
            }
        }
    }
    """

    COMMAND_MESSAGES = (
        # Keep these together as they're bound to function keys and destined
        # for the footer.
        Help,
        TogglePEPDetails,
        Quit,
        RedownloadPEPs,
        # Everything else.
        ChangeTheme,
        Escape,
        FindPEP,
        Search,
        SearchAuthor,
        SearchPythonVersion,
        SearchStatus,
        SearchType,
        ShowAll,
        SortByCreated,
        SortByNumber,
        SortByTitle,
        ToggleAuthorsSortOrder,
        TogglePythonVersionsSortOrder,
        ToggleStatusesSortOrder,
        ToggleTypesSortOrder,
    )

    BINDINGS = Command.bindings(*COMMAND_MESSAGES)

    COMMANDS = {MainCommands}

    all_peps: var[PEPs] = var(PEPs)
    """All the PEPs that we know about."""

    active_peps: var[PEPs] = var(PEPs)
    """The currently-active set of PEPs."""

    selected_pep: var[PEP | None] = var(None)
    """The currently-selected PEP."""

    def compose(self) -> ComposeResult:
        """Compose the content of the main screen."""
        yield Header()
        with Horizontal():
            yield Navigation(load_configuration(), classes="panel").data_bind(
                Main.all_peps, Main.active_peps
            )
            yield PEPsView(classes="panel").data_bind(Main.active_peps)
            yield PEPDetails(classes="panel").data_bind(pep=Main.selected_pep)
        yield Footer()

    @dataclass
    class Loaded(Message):
        """A message sent when PEP data is loaded."""

        peps: PEPs
        """The PEP data that was loaded."""

    @work(thread=True)
    def load_pep_data(self) -> None:
        """Load the local copy of the PEP data."""
        try:
            self.post_message(
                self.Loaded(
                    PEPs(
                        PEP.from_json(pep)
                        for pep in loads(pep_data().read_text()).values()
                    )
                )
            )
        except IOError as error:
            self.notify(str(error), title="Error loading PEP data", severity="error")

    @work(thread=True)
    async def download_pep_data(self) -> None:
        """Download a fresh copy of the PEP data."""
        self.notify("Downloading PEPs from the API...")
        peps, raw_data = await API().get_peps()
        try:
            pep_data().write_text(dumps(raw_data, indent=4), encoding="utf-8")
        except IOError as error:
            self.notify(str(error), title="Error saving PEP data", severity="error")
        self.post_message(self.Loaded(PEPs(peps)))

    @on(Loaded)
    def load_fresh_peps(self, message: Loaded) -> None:
        """React to a fresh set of PEPs being made available."""
        self.all_peps = message.peps.sorted_by(load_configuration().peps_sort_order)

    def on_mount(self) -> None:
        """Configure the application once the DOM is mounted."""
        self.set_class(load_configuration().details_visble, "details-visible")
        if pep_data().exists():
            self.load_pep_data()
        else:
            self.download_pep_data()

    def watch_all_peps(self) -> None:
        """React to the full set of PEPs being updated."""
        self.active_peps = self.all_peps
        PEPsCommands.peps = self.all_peps

    def watch_active_peps(self) -> None:
        """React to the active PEPs being updated."""
        self.sub_title = f"{self.active_peps.description} ({len(self.active_peps)})"
        AuthorCommands.active_peps = self.active_peps
        PythonVersionCommands.active_peps = self.active_peps
        StatusCommands.active_peps = self.active_peps
        TypeCommands.active_peps = self.active_peps

    def _show_palette(self, provider: type[CommandsProvider]) -> None:
        """Show a particular command palette.

        Args:
            provider: The commands provider for the palette.
        """
        self.app.push_screen(
            CommandPalette(
                providers=(provider,),
                placeholder=provider.prompt(),
            )
        )

    @on(PEPsView.PEPHighlighted)
    def select_pep(self, message: PEPsView.PEPHighlighted) -> None:
        """Make the currently-selected PEP the one to view."""
        self.selected_pep = message.pep

    @on(PEPsView.Empty)
    def deselect_pep(self) -> None:
        """Handle there being no highlighted PEP."""
        self.selected_pep = None

    @on(ShowAll)
    def action_show_all_command(self) -> None:
        """Show all PEPs."""
        self.active_peps = self.all_peps

    @on(ShowType)
    def show_type(self, command: ShowType) -> None:
        """Filter the PEPs by a given type.

        Args:
            command: The command requesting the filter.
        """
        self.active_peps &= WithType(command.type)

    @on(ShowStatus)
    def show_status(self, command: ShowStatus) -> None:
        """Filter the PEPs by a given status.

        Args:
            command: The command requesting the filter.
        """
        self.active_peps &= WithStatus(command.status)

    @on(ShowPythonVersion)
    def action_show_python_version(self, command: ShowPythonVersion) -> None:
        """Filter the PEPs by a given Python version.

        Args:
            command: The command requesting the filter.
        """
        self.active_peps &= WithPythonVersion(command.version)

    @on(ShowAuthor)
    def show_author(self, command: ShowAuthor) -> None:
        """Filter the PEPs by a given author.

        Args:
            command: The command requesting the filter.
        """
        self.active_peps &= WithAuthor(command.author)

    @on(GotoPEP)
    def goto_pep(self, command: GotoPEP) -> None:
        """Visit a specific PEP by its number.

        Args:
            command: The command requesting the PEP by number.
        """
        if command.number in self.active_peps:
            self.query_one(PEPsView).goto_pep(command.number)
        elif command.number in self.all_peps:
            self.notify(
                f"PEP{command.number} wasn't in the active filter; switching to all PEPs...",
                severity="warning",
            )
            self.active_peps = self.all_peps
            self.call_after_refresh(self.query_one(PEPsView).goto_pep, command.number)
        else:
            self.notify(
                f"PEP{command.number} doesn't exist. Perhaps you'll be the one to write it?",
                title="No such PEP",
                severity="error",
            )

    @on(VisitPEP)
    def visit_pep(self, command: VisitPEP) -> None:
        """Visit a given PEP's webpage.

        Args:
            command: The command requesting the visit.
        """
        if command.pep.url:
            visit_url(command.pep.url)
        else:
            self.notify(f"PEP{command.pep.number} has no associated URL")

    @on(FindPEP)
    def action_find_pep_command(self) -> None:
        """Find a PEP and jump to it."""
        self._show_palette(PEPsCommands)

    @on(Search)
    @work
    async def action_search_command(self) -> None:
        """Free-text search within the PEPs."""
        if search_text := await self.app.push_screen_wait(SearchInput()):
            self.active_peps = self.active_peps & Containing(search_text)

    @on(SearchAuthor)
    def action_search_author_command(self) -> None:
        """Search for an author and use them as a filter."""
        self._show_palette(AuthorCommands)

    @on(SearchPythonVersion)
    def action_search_python_version_command(self) -> None:
        """Search for a Python version and then use it as a filter."""
        self._show_palette(PythonVersionCommands)

    @on(SearchStatus)
    def action_search_status_command(self) -> None:
        """Search for a status and use it as a filter."""
        self._show_palette(StatusCommands)

    @on(SearchType)
    def action_search_type_command(self) -> None:
        """Search for a PEP type and then use it as a filter."""
        self._show_palette(TypeCommands)

    @on(RedownloadPEPs)
    def action_redownload_peps_command(self) -> None:
        """Redownload PEPs from the API."""
        self.download_pep_data()

    @on(Help)
    def action_help_command(self) -> None:
        """Toggle the display of the help panel."""
        self.app.push_screen(HelpScreen(self))

    @on(ChangeTheme)
    def action_change_theme_command(self) -> None:
        """Show the theme picker."""
        self.app.search_themes()

    @on(Quit)
    def action_quit_command(self) -> None:
        """Quit the application."""
        self.app.exit()

    @on(Escape)
    def action_escape_command(self) -> None:
        """Handle escaping.

        The action's approach is to step-by-step back out from the 'deepest'
        level to the topmost, and if we're at the topmost then exit the
        application.
        """
        if self.focused == self.query_one(PEPsView):
            self.set_focus(self.query_one(Navigation))
        elif (
            self.focused
            and self.query_one(PEPDetails) in self.focused.ancestors_with_self
        ):
            self.set_focus(self.query_one(PEPsView))
        else:
            self.app.exit()

    @on(TogglePEPDetails)
    def action_toggle_pep_details_command(self) -> None:
        """Toggle the display of the PEP details panel."""
        self.toggle_class("details-visible")
        with update_configuration() as config:
            config.details_visble = self.has_class("details-visible")

    @on(ToggleTypesSortOrder)
    def action_toggle_types_sort_order_command(self) -> None:
        """Toggle the sort order of the types."""
        with update_configuration() as config:
            config.sort_types_by_count = not config.sort_types_by_count
            self.query_one(Navigation).sort_types_by_count = config.sort_types_by_count

    @on(ToggleStatusesSortOrder)
    def action_toggle_statuses_sort_order_command(self) -> None:
        """Toggle the sort order of the statuses."""
        with update_configuration() as config:
            config.sort_statuses_by_count = not config.sort_statuses_by_count
            self.query_one(
                Navigation
            ).sort_statuses_by_count = config.sort_statuses_by_count

    @on(TogglePythonVersionsSortOrder)
    def action_toggle_python_versions_sort_order_command(self) -> None:
        """Toggle the sort order of the Python Versions."""
        with update_configuration() as config:
            config.sort_python_versions_by_count = (
                not config.sort_python_versions_by_count
            )
            self.query_one(
                Navigation
            ).sort_python_versions_by_count = config.sort_python_versions_by_count

    @on(ToggleAuthorsSortOrder)
    def action_toggle_authors_sort_order_command(self) -> None:
        """Toggle the sort order of the authors."""
        with update_configuration() as config:
            config.sort_authors_by_count = not config.sort_authors_by_count
            self.query_one(
                Navigation
            ).sort_authors_by_count = config.sort_authors_by_count

    @on(SortByCreated)
    def action_sort_by_created_command(self) -> None:
        """Sort the PEPs by their date created."""
        with update_configuration() as config:
            config.peps_sort_order = "created"
            self.active_peps = self.active_peps.sorted_by(config.peps_sort_order)

    @on(SortByNumber)
    def action_sort_by_number_command(self) -> None:
        """Sort the PEPs by their number."""
        with update_configuration() as config:
            config.peps_sort_order = "number"
            self.active_peps = self.active_peps.sorted_by(config.peps_sort_order)

    @on(SortByTitle)
    def action_sort_by_title_command(self) -> None:
        """Sort the PEPs by their title."""
        with update_configuration() as config:
            config.peps_sort_order = "title"
            self.active_peps = self.active_peps.sorted_by(config.peps_sort_order)


### main.py ends here



"""Provides messages that apply to the whole application."""

##############################################################################
# Local imports.
from .main import (
    GotoPEP,
    ShowAuthor,
    ShowPythonVersion,
    ShowStatus,
    ShowType,
    VisitPEP,
)

##############################################################################
# Exports.
__all__ = [
    "GotoPEP",
    "ShowAuthor",
    "ShowPythonVersion",
    "ShowStatus",
    "ShowType",
    "VisitPEP",
]

### __init__.py ends here



"""Commands for affecting navigation sort ordering."""

##############################################################################
# Local imports.
from .base import Command


##############################################################################
class ToggleTypesSortOrder(Command):
    """Toggle the sort order of types in the navigation panel"""

    BINDING_KEY = "T"
    SHOW_IN_FOOTER = False


##############################################################################
class ToggleStatusesSortOrder(Command):
    """Toggle the sort order of the statuses in the navigation panel"""

    BINDING_KEY = "S"
    SHOW_IN_FOOTER = False


##############################################################################
class TogglePythonVersionsSortOrder(Command):
    """Toggle the sort order of Python versions in the navigation panel"""

    BINDING_KEY = "V"
    SHOW_IN_FOOTER = False


##############################################################################
class ToggleAuthorsSortOrder(Command):
    """Toggle the sort order of the authors in the navigation panel"""

    BINDING_KEY = "A"
    SHOW_IN_FOOTER = False


### navigation_sorting.py ends here



"""Provides a PEP API client class."""

##############################################################################
# Python imports.
from ssl import SSLCertVerificationError
from typing import Any, Final

##############################################################################
# HTTPX imports.
from httpx import AsyncClient, HTTPStatusError, RequestError

##############################################################################
# Local imports.
from .pep import PEP


##############################################################################
class API:
    """API client for peps.python.org."""

    AGENT: Final[str] = "Peplum (https://github.com/davep/peplum)"
    """The agent string to use when talking to the API."""

    _URL: Final[str] = "https://peps.python.org/api/peps.json"
    """The URL of the PEP download API."""

    class Error(Exception):
        """Base class for Raindrop errors."""

    class RequestError(Error):
        """Exception raised if there was a problem making an API request."""

    def __init__(self) -> None:
        """Initialise the client object."""
        self._client_: AsyncClient | None = None
        """The internal reference to the HTTPX client."""

    @property
    def _client(self) -> AsyncClient:
        """The HTTPX client."""
        if self._client_ is None:
            self._client_ = AsyncClient()
        return self._client_

    async def get_peps(self) -> tuple[list[PEP], dict[int, dict[str, Any]]]:
        """Download a fresh list of all known PEPs.

        Returns:
            A tuple of a list of all known PEPs and the raw JSON they were
            created from.
        """
        try:
            response = await self._client.get(
                self._URL, headers={"user-agent": self.AGENT}
            )
        except (RequestError, SSLCertVerificationError) as error:
            raise self.RequestError(str(error)) from None

        try:
            response.raise_for_status()
        except HTTPStatusError as error:
            raise self.RequestError(str(error)) from None

        return [PEP.from_json(pep) for pep in response.json().values()], response.json()


### api.py ends here



"""Provides the main application class."""

##############################################################################
# Textual imports.
from textual.app import App, InvalidThemeError
from textual.binding import Binding

##############################################################################
# Local imports.
from .data import (
    load_configuration,
    update_configuration,
)
from .screens import Main


##############################################################################
class Peplum(App[None]):
    """The main application class."""

    CSS = """
    CommandPalette > Vertical {
        width: 75%; /* Full-width command palette looks like garbage. Fix that. */
        background: $panel;
        SearchIcon {
            display: none;
        }
        OptionList {
            /* Make the scrollbar less gross. */
            scrollbar-background: $panel;
            scrollbar-background-hover: $panel;
            scrollbar-background-active: $panel;
        }
    }

    /* Remove cruft from the Header. */
    Header {
        /* The header icon is ugly and pointless. Remove it. */
        HeaderIcon {
            visibility: hidden;
        }

        /* The tall version of the header is utterly useless. Nuke that. */
        &.-tall {
            height: 1 !important;
        }
    }

    /* General style tweaks that affect all widgets. */
    * {
        /* Let's make scrollbars a wee bit thinner. */
        scrollbar-size-vertical: 1;
    }
    """

    BINDINGS = [
        Binding(
            "ctrl+p, super+x, :",
            "command_palette",
            "Commands",
            show=False,
            tooltip="Show the command palette",
        ),
    ]

    COMMANDS = set()

    def __init__(self) -> None:
        """Initialise the application."""
        super().__init__()
        configuration = load_configuration()
        if configuration.theme is not None:
            try:
                self.theme = configuration.theme
            except InvalidThemeError:
                pass

    def watch_theme(self) -> None:
        """Save the application's theme when it's changed."""
        with update_configuration() as config:
            config.theme = self.theme

    def on_mount(self) -> None:
        """Display the main screen."""
        self.push_screen(Main())


### peplum.py ends here



"""Tests for the class that holds and manages PEPs."""

##############################################################################
# Python imports.
from typing import Final, get_args

##############################################################################
# Pytest imports.
from pytest import mark

##############################################################################
# Local imports.
from peplum.app.data import (
    Containing,
    PEPs,
    WithAuthor,
    WithPythonVersion,
    WithStatus,
    WithType,
)
from peplum.app.data.peps import Filter
from peplum.peps import PEP, PEPStatus, PEPType

##############################################################################
SAMPLE_PEPS: Final[tuple[PEP, ...]] = (
    PEP.from_json(
        {
            "number": 1,
            "title": "PEP Purpose and Guidelines",
            "authors": "Author 1, Author 2, Author 3",
            "discussions_to": None,
            "status": "Active",
            "type": "Process",
            "topic": "",
            "created": "13-Jun-2000",
            "python_version": "3.7, 3.8",
            "post_history": "21-Mar-2001, 29-Jul-2002, 03-May-2003, 05-May-2012, 07-Apr-2013",
            "resolution": None,
            "requires": None,
            "replaces": None,
            "superseded_by": None,
            "url": "https://peps.python.org/pep-0001/",
        }
    ),
    PEP.from_json(
        {
            "number": 458,
            "title": "Secure PyPI downloads with signed repository metadata",
            "authors": "Author 1",
            "discussions_to": "https://discuss.python.org/t/pep-458-secure-pypi-downloads-with-package-signing/2648",
            "status": "Accepted",
            "type": "Standards Track",
            "topic": "packaging",
            "created": "27-Sep-2013",
            "python_version": None,
            "post_history": "06-Jan-2019, 13-Nov-2019",
            "resolution": "https://discuss.python.org/t/pep-458-secure-pypi-downloads-with-package-signing/2648/115",
            "requires": None,
            "replaces": None,
            "superseded_by": None,
            "url": "https://peps.python.org/pep-0458/",
        }
    ),
    PEP.from_json(
        {
            "number": 467,
            "title": "Minor API improvements for binary sequences",
            "authors": "Author 1, Jr.",
            "discussions_to": "https://discuss.python.org/t/42001",
            "status": "Draft",
            "type": "Informational",
            "topic": "",
            "created": "30-Mar-2014",
            "python_version": "3.13",
            "post_history": "30-Mar-2014, 15-Aug-2014, 16-Aug-2014, 07-Jun-2016, 01-Sep-2016, 13-Apr-2021, 03-Nov-2021, 27-Dec-2023",
            "resolution": None,
            "requires": None,
            "replaces": None,
            "superseded_by": None,
            "url": "https://peps.python.org/pep-0467/",
        }
    ),
    PEP.from_json(
        {
            "number": 639,
            "title": "Improving License Clarity with Better Package Metadata",
            "authors": "Author 1, Author 1, Jr.",
            "discussions_to": "https://discuss.python.org/t/53020",
            "status": "Provisional",
            "type": "Process",
            "topic": "packaging",
            "created": "15-Aug-2019",
            "python_version": None,
            "post_history": "`15-Aug-2019 <https://discuss.python.org/t/2154>`__, `17-Dec-2021 <https://discuss.python.org/t/12622>`__, `10-May-2024 <https://discuss.python.org/t/53020>`__,",
            "resolution": "https://discuss.python.org/t/53020/106",
            "requires": None,
            "replaces": None,
            "superseded_by": None,
            "url": "https://peps.python.org/pep-0639/",
        }
    ),
    PEP.from_json(
        {
            "number": 213,
            "title": "Attribute Access Handlers",
            "authors": "Author 1, Jr., Author 1",
            "discussions_to": None,
            "status": "Deferred",
            "type": "Standards Track",
            "topic": "",
            "created": "21-Jul-2000",
            "python_version": "2.1",
            "post_history": None,
            "resolution": None,
            "requires": None,
            "replaces": None,
            "superseded_by": None,
            "url": "https://peps.python.org/pep-0213/",
        }
    ),
    PEP.from_json(
        {
            "number": 204,
            "title": "Range Literals",
            "authors": "Author 2",
            "discussions_to": None,
            "status": "Rejected",
            "type": "Informational",
            "topic": "",
            "created": "14-Jul-2000",
            "python_version": "2.0",
            "post_history": None,
            "resolution": None,
            "requires": None,
            "replaces": None,
            "superseded_by": None,
            "url": "https://peps.python.org/pep-0204/",
        }
    ),
    PEP.from_json(
        {
            "number": 3,
            "title": "Guidelines for Handling Bug Reports",
            "authors": "Author 1",
            "discussions_to": None,
            "status": "Withdrawn",
            "type": "Process",
            "topic": "",
            "created": "25-Sep-2000",
            "python_version": None,
            "post_history": None,
            "resolution": None,
            "requires": None,
            "replaces": None,
            "superseded_by": None,
            "url": "https://peps.python.org/pep-0003/",
        }
    ),
    PEP.from_json(
        {
            "number": 100,
            "title": "Python Unicode Integration",
            "authors": "Author 1",
            "discussions_to": None,
            "status": "Final",
            "type": "Standards Track",
            "topic": "",
            "created": "10-Mar-2000",
            "python_version": "2.0",
            "post_history": None,
            "resolution": None,
            "requires": None,
            "replaces": None,
            "superseded_by": None,
            "url": "https://peps.python.org/pep-0100/",
        }
    ),
    PEP.from_json(
        {
            "number": 5,
            "title": "Guidelines for Language Evolution",
            "authors": "Author 3",
            "discussions_to": None,
            "status": "Superseded",
            "type": "Informational",
            "topic": "",
            "created": "26-Oct-2000",
            "python_version": None,
            "post_history": None,
            "resolution": None,
            "requires": None,
            "replaces": None,
            "superseded_by": "387",
            "url": "https://peps.python.org/pep-0005/",
        }
    ),
)
"""Some sample PEP data to work off."""


##############################################################################
def test_no_peps() -> None:
    """An empty PEPs object sound have no length."""
    assert len(PEPs()) == 0


##############################################################################
def test_has_peps() -> None:
    """A non-empty PEPs object should report the correct length."""
    assert len(PEPs(SAMPLE_PEPS)) == len(SAMPLE_PEPS)


##############################################################################
def test_status_counts() -> None:
    """We should be able to count the statuses."""
    assert {(status.status, status.count) for status in PEPs(SAMPLE_PEPS).statuses} == {
        (status, 1) for status in get_args(PEPStatus)
    }


##############################################################################
def test_type_counts() -> None:
    """We should be able to count the types."""
    assert {
        (pep_type.type, pep_type.count) for pep_type in PEPs(SAMPLE_PEPS).types
    } == {(pep_type, 3) for pep_type in get_args(PEPType)}


##############################################################################
def test_python_version_counts() -> None:
    """We should be able to count the Python versions."""
    assert {
        (version.version, version.count)
        for version in PEPs(SAMPLE_PEPS).python_versions
    } == {
        ("", 4),
        ("2.0", 2),
        ("2.1", 1),
        ("3.7", 1),
        ("3.8", 1),
        ("3.13", 1),
    }


##############################################################################
def test_author_counts() -> None:
    """We should be able to count the authors."""
    assert {(author.author, author.count) for author in PEPs(SAMPLE_PEPS).authors} == {
        ("Author 1", 6),
        ("Author 1, Jr.", 3),
        ("Author 2", 2),
        ("Author 3", 2),
    }


##############################################################################
@mark.parametrize(
    "pep_filter, expected",
    (
        (Containing("Unicode"), 1),
        (Containing("unicode"), 1),
        (Containing("UNICODE"), 1),
        (Containing("UNICOD"), 1),
        (Containing("NiCOd"), 1),
        (Containing("author 1"), 7),
        (Containing("JR"), 3),
        (Containing("STANDARDS TRACK"), 3),
        (Containing("https://peps.python.org/pep-0467/"), 1),
        (WithAuthor("Author 1, Jr."), 3),
        (WithAuthor("Author 1, Jr"), 0),
        (WithAuthor("author 1"), 6),
        (WithAuthor("AUTHOR 1, JR."), 3),
        (WithAuthor("JR"), 0),
        (WithPythonVersion("3.13"), 1),
        (WithPythonVersion(""), 4),
        (WithStatus("Accepted"), 1),
        (WithType("Standards Track"), 3),
    ),
)
def test_filter(pep_filter: Filter, expected: int) -> None:
    """Test that we can filter a collection of PEPs."""
    assert len(PEPs(SAMPLE_PEPS) & pep_filter) == expected


### test_peps.py ends here



"""Provides functions and classes for managing the app's data."""

##############################################################################
# Local imports.
from .config import (
    Configuration,
    load_configuration,
    save_configuration,
    update_configuration,
)
from .peps import (
    AuthorCount,
    Containing,
    PEPCount,
    PEPs,
    PythonVersionCount,
    StatusCount,
    TypeCount,
    WithAuthor,
    WithPythonVersion,
    WithStatus,
    WithType,
    pep_data,
)

##############################################################################
# Exports.
__all__ = [
    "pep_data",
    "AuthorCount",
    "Configuration",
    "Containing",
    "load_configuration",
    "PEPCount",
    "PEPs",
    "PythonVersionCount",
    "save_configuration",
    "StatusCount",
    "TypeCount",
    "update_configuration",
    "WithAuthor",
    "WithPythonVersion",
    "WithStatus",
    "WithType",
]

### __init__.py ends here



"""PEP type filtering commands for the command palette."""

##############################################################################
# Local imports.
from ..data import PEPs
from ..messages import ShowType
from .commands_provider import CommandHit, CommandHits, CommandsProvider


##############################################################################
class TypeCommands(CommandsProvider):
    """A command palette provider related to types."""

    active_peps: PEPs | None = None
    """The currently-active collection of PEPs to get the types of."""

    @classmethod
    def prompt(cls) -> str:
        """The prompt for the command provider."""
        return (
            "Also search for PEPs of type..."
            if cls.active_peps and cls.active_peps.is_filtered
            else "Search for PEPs of type..."
        )

    def commands(self) -> CommandHits:
        """Provide the type-based command data for the command palette.

        Yields:
            The commands for the command palette.
        """
        if self.active_peps is None:
            return
        help_prefix = "Also filter" if self.active_peps.is_filtered else "Filter"
        command_prefix = "Also of type" if self.active_peps.is_filtered else "Of type"
        for pep_type in self.active_peps.types:
            yield CommandHit(
                f"{command_prefix} {pep_type.type}",
                f"{help_prefix} to PEPs of type {pep_type.type} (narrows down to {pep_type.count})",
                ShowType(pep_type.type),
            )


### types.py ends here



"""Provides command-oriented messages for the application.

These messages differ a little from other messages in that they have a
common base class and provide information such as help text, binding
information, etc.
"""

##############################################################################
# Local imports.
from .base import Command
from .filtering import (
    Search,
    SearchAuthor,
    SearchPythonVersion,
    SearchStatus,
    SearchType,
    ShowAll,
)
from .finding import FindPEP
from .main import ChangeTheme, Escape, Help, Quit, RedownloadPEPs, TogglePEPDetails
from .navigation_sorting import (
    ToggleAuthorsSortOrder,
    TogglePythonVersionsSortOrder,
    ToggleStatusesSortOrder,
    ToggleTypesSortOrder,
)
from .peps_sorting import SortByCreated, SortByNumber, SortByTitle

##############################################################################
# Exports.
__all__ = [
    "Command",
    "ChangeTheme",
    "Escape",
    "FindPEP",
    "Help",
    "Quit",
    "RedownloadPEPs",
    "Search",
    "SearchAuthor",
    "SearchPythonVersion",
    "SearchStatus",
    "SearchType",
    "ShowAll",
    "SortByCreated",
    "SortByNumber",
    "SortByTitle",
    "ToggleAuthorsSortOrder",
    "TogglePEPDetails",
    "TogglePythonVersionsSortOrder",
    "ToggleStatusesSortOrder",
    "ToggleTypesSortOrder",
]

### __init__.py ends here



"""Provides code for handling a collection of PEPs."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from collections import Counter
from dataclasses import dataclass
from functools import total_ordering
from itertools import chain
from operator import attrgetter
from pathlib import Path
from typing import Iterable, Iterator, Literal, TypeAlias

##############################################################################
# Packaging imports.
from packaging.version import InvalidVersion, Version

##############################################################################
# Local imports.
from ...peps import PEP, PEPStatus, PEPType
from .locations import data_dir


##############################################################################
def pep_data() -> Path:
    """The path to the local copy of the PEP data."""
    return data_dir() / "peps.json"


##############################################################################
@dataclass(frozen=True)
@total_ordering
class StatusCount:
    """Holds a count of a particular PEP status."""

    status: PEPStatus
    """The PEP status."""
    count: int
    """The count."""

    def __gt__(self, value: object, /) -> bool:
        if isinstance(value, StatusCount):
            return self.status > value.status
        raise NotImplementedError

    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, StatusCount):
            return self.status == value.status
        raise NotImplementedError


##############################################################################
@dataclass(frozen=True)
@total_ordering
class TypeCount:
    """Holds a count of a particular PEP type."""

    type: PEPType
    """The PEP type."""
    count: int
    """The count."""

    def __gt__(self, value: object, /) -> bool:
        if isinstance(value, TypeCount):
            return self.type > value.type
        raise NotImplementedError

    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, TypeCount):
            return self.type == value.type
        raise NotImplementedError


##############################################################################
@dataclass(frozen=True)
@total_ordering
class PythonVersionCount:
    """Holds a count of a particular PEP python version."""

    version: str
    """The Python version."""
    count: int
    """The count."""

    def __gt__(self, value: object, /) -> bool:
        if isinstance(value, PythonVersionCount):
            try:
                return Version(self.version) > Version(value.version)
            except InvalidVersion:
                return self.version > value.version
        raise NotImplementedError

    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, PythonVersionCount):
            try:
                return Version(self.version) == Version(value.version)
            except InvalidVersion:
                return self.version == value.version
        raise NotImplementedError


##############################################################################
@dataclass(frozen=True)
@total_ordering
class AuthorCount:
    """Holds a count of a particular PEP author."""

    author: str
    """The PEP author."""
    count: int
    """The count."""

    def __gt__(self, value: object, /) -> bool:
        if isinstance(value, AuthorCount):
            return self.author.casefold() > value.author.casefold()
        raise NotImplementedError

    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, AuthorCount):
            return self.author.casefold() == value.author.casefold()
        raise NotImplementedError


##############################################################################
PEPCount: TypeAlias = StatusCount | TypeCount | PythonVersionCount | AuthorCount
"""The type of the various counts."""

##############################################################################
Filters: TypeAlias = tuple["Filter", ...]
"""The type of a collection of filters."""


##############################################################################
class Filter:
    """Base class for the raindrop filters."""

    def __rand__(self, _: PEP) -> bool:
        return False

    def __radd__(self, filters: Filters) -> Filters:
        return (*filters, self)

    def __eq__(self, value: object) -> bool:
        if isinstance(value, self.__class__):
            return str(value) == str(self)
        return False


##############################################################################
class WithStatus(Filter):
    """Filter on a PEP's status."""

    def __init__(self, status: PEPStatus) -> None:
        """Initialise the object.

        Args:
            status: The status to filter on.
        """
        self._status = status
        """The status to filter on."""

    def __rand__(self, pep: PEP) -> bool:
        return pep.status == self._status

    def __str__(self) -> str:
        return str(self._status)


##############################################################################
class WithType(Filter):
    """Filter on a PEP's status."""

    def __init__(self, pep_type: PEPType) -> None:
        """Initialise the object.

        Args:
            pep_type: The type to filter on.
        """
        self._type = pep_type
        """The type to filter on."""

    def __rand__(self, pep: PEP) -> bool:
        return pep.type == self._type

    def __str__(self) -> str:
        return str(self._type)


##############################################################################
class WithPythonVersion(Filter):
    """Filter on a PEP's Python version."""

    def __init__(self, version: str) -> None:
        """Initialise the object.

        Args:
            version: The version to filter on.
        """
        self._version = version
        """The type to filter on."""

    def __rand__(self, pep: PEP) -> bool:
        return (
            self._version in pep.python_version
            if self._version
            else not pep.python_version
        )

    def __str__(self) -> str:
        return self._version or "None"


##############################################################################
class WithAuthor(Filter):
    """Filter on a PEP's author."""

    def __init__(self, author: str) -> None:
        """Initialise the object.

        Args:
            author: The author to filter on.
        """
        self._author = author
        self._folded_author = author.casefold()
        """The author to filter on."""

    def __rand__(self, pep: PEP) -> bool:
        return self._folded_author in (author.casefold() for author in pep.authors)

    def __str__(self) -> str:
        return str(self._author)


##############################################################################
class Containing(Filter):
    """Filter on text found within a PEP's data."""

    def __init__(self, text: str) -> None:
        """Initialise the object.

        Args:
            text: The text to filter on.
        """
        self._text = text
        """The text to look for."""

    def __rand__(self, pep: PEP) -> bool:
        return self._text in pep

    def __str__(self) -> str:
        return self._text

    def __eq__(self, value: object) -> bool:
        if isinstance(value, Containing):
            return str(value).casefold() == self._text.casefold()
        return super().__eq__(value)


##############################################################################
SortOrder: TypeAlias = Literal["number", "created", "title"]
"""Sort orders for PEPs."""


##############################################################################
class PEPs:
    """Class that holds a collection of PEPs."""

    def __init__(
        self,
        peps: Iterable[PEP] | None = None,
        filters: Filters | None = None,
        sort_order: SortOrder = "number",
    ) -> None:
        """Initialise the object.

        Args:
            peps: The PEPs to hold.
            filters: The filters that got to this set of PEPs.
        """
        self._peps: dict[int, PEP] = (
            {} if peps is None else {pep.number: pep for pep in peps}
        )
        """The PEPs."""
        self._filters = () if filters is None else filters
        """The filters that got to this set of PEPs."""
        self._sort_order: SortOrder = sort_order
        """The sort order for the PEPs."""

    @property
    def is_filtered(self) -> bool:
        """Does this collection of PEPs have a filter?"""
        return bool(self._filters)

    @property
    def statuses(self) -> tuple[StatusCount, ...]:
        """The status and their counts as found in the PEPs."""
        return tuple(
            StatusCount(status, count)
            for status, count in Counter[PEPStatus](pep.status for pep in self).items()
        )

    @property
    def types(self) -> tuple[TypeCount, ...]:
        """The types and their counts as found in the PEPs."""
        return tuple(
            TypeCount(pep_type, count)
            for pep_type, count in Counter[PEPType](pep.type for pep in self).items()
        )

    @property
    def python_versions(self) -> tuple[PythonVersionCount, ...]:
        """The Python versions and their counts as found in the PEPs.

        Notes:
            A count for an empty string is included, this is the count of
            PEPs that have no Python version associated with them.
        """
        return tuple(
            PythonVersionCount(version, count)
            for version, count in Counter(
                chain(*(pep.python_version or ("",) for pep in self))
            ).items()
        )

    @property
    def authors(self) -> tuple[AuthorCount, ...]:
        """The authors and their counts as found in the PEPs."""
        return tuple(
            AuthorCount(author, count)
            for author, count in Counter(chain(*(pep.authors for pep in self))).items()
        )

    def _describe(self, name: str, filter_type: type[Filter]) -> str | None:
        """Describe the user's use of a particular filter.

        Args:
            name: The name to give the filter.
            filter_type: The type of filter to look for.

        Returns:
            A description if the filter is used, or `None`.
        """
        return (
            f"{name} {' and '.join(filters)}"
            if (
                filters := [
                    f"{candidate}"
                    for candidate in self._filters
                    if isinstance(candidate, filter_type)
                ]
            )
            else None
        )

    @property
    def description(self) -> str:
        """The description of the content of the PEPs collection."""
        filters = [
            candidate
            for candidate in [
                self._describe(name, filter_type)
                for name, filter_type in (
                    ("Containing", Containing),
                    ("Type", WithType),
                    ("Status", WithStatus),
                    ("Version", WithPythonVersion),
                    ("Author", WithAuthor),
                )
            ]
            if candidate
        ] or ["All"]

        match self._sort_order:
            case "number":
                sort_order = "PEP Number"
            case "created":
                sort_order = "Date Created"
            case "title":
                sort_order = "PEP Title"

        return "; ".join(filters + ([f"Sorted by {sort_order}"] if sort_order else []))

    def __and__(self, new_filter: Filter) -> PEPs:
        """Get the PEPs match a given filter.

        Args:
            new_filter: The new filter to apply.

        Returns:
            The subset of PEPs that match the given filter.
        """
        # Only return a new PEPs object if the filter we've been given isn't
        # already active.
        return (
            self
            if new_filter in self._filters
            else PEPs(
                (pep for pep in self if pep & new_filter),
                self._filters + new_filter,
                self._sort_order,
            )
        )

    def sorted_by(self, sort_order: SortOrder) -> PEPs:
        """Get the PEPs sorted in a particular way.

        Args:
            The sort order.

        Returns:
            The PEPs sorted in the required way.
        """
        return PEPs(self, self._filters, sort_order)

    def __contains__(self, pep: PEP | int) -> bool:
        """Is the given PEP in here?"""
        return (pep.number if isinstance(pep, PEP) else pep) in self._peps

    def __iter__(self) -> Iterator[PEP]:
        """The object as an iterator."""
        return iter(sorted(self._peps.values(), key=attrgetter(self._sort_order)))

    def __len__(self) -> int:
        """The count of PEPs in the object."""
        return len(self._peps)


### peps.py ends here



"""Provides the main application commands for the command palette."""

##############################################################################
# Local imports.
from ..commands import (
    ChangeTheme,
    Escape,
    FindPEP,
    Help,
    Quit,
    RedownloadPEPs,
    Search,
    SearchAuthor,
    SearchPythonVersion,
    SearchStatus,
    SearchType,
    ShowAll,
    SortByCreated,
    SortByNumber,
    SortByTitle,
    ToggleAuthorsSortOrder,
    TogglePEPDetails,
    TogglePythonVersionsSortOrder,
    ToggleStatusesSortOrder,
    ToggleTypesSortOrder,
)
from .commands_provider import CommandHits, CommandsProvider


##############################################################################
class MainCommands(CommandsProvider):
    """Provides some top-level commands for the application."""

    def commands(self) -> CommandHits:
        """Provide the main application commands for the command palette.

        Yields:
            The commands for the command palette.
        """
        yield ChangeTheme()
        yield Escape()
        yield FindPEP()
        yield Help()
        yield Quit()
        yield RedownloadPEPs()
        yield Search()
        yield SearchAuthor()
        yield SearchPythonVersion()
        yield SearchStatus()
        yield SearchType()
        yield ShowAll()
        yield SortByCreated()
        yield SortByNumber()
        yield SortByTitle()
        yield ToggleAuthorsSortOrder()
        yield TogglePEPDetails()
        yield TogglePythonVersionsSortOrder()
        yield ToggleStatusesSortOrder()
        yield ToggleTypesSortOrder()


### main.py ends here
